////////////////////////////////////////////////////////////////////////////////
//
// (C) 2010 Sarah Eris Horsley Caffee
//
// mandelbrot - A simple Mandelbrot Set viewer.
//
// App.cpp
// Application class implementation
//
////////////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <vector>

#include <dirent.h>

#include "png.h"
#include "SDL.h"

#include "App.h"
#include "Options.h"

#include "Mandelbrot.h"
#include "Julia.h"


#define USE_SDL_WAITEVENT

std::vector<App::view *> App::old_views;

////////////////////////////////////////////////////////////////////////////////

App::App() :
   running (true),
   redraw (false),
   surf_display (NULL),
   surf_selection (NULL),
   surf_fractal (NULL),
   selection_x (-1),
   selection_y (-1),
   selection_width (-1),
   setting_zoom (false),
   can_set_julia_k (false),
   setting_julia_k (false)
   {
   }

////////////////////////////////////////////////////////////////////////////////
App::~App()
   {
   if (old_views.size() > 0)
      {
      view * v = old_views.back();
      old_views.pop_back();
      delete v;
      }
   }

////////////////////////////////////////////////////////////////////////////////

int App::run()
   {
   if (init() == false)
      return 1;

   SDL_Event event;

   while (running)
      {
#ifdef USE_SDL_WAITEVENT
      SDL_WaitEvent(&event);
      do
	 {
	 on_event(&event);
	 }
      while (SDL_PollEvent(&event));
#else
      while (wait_event_timeout(&event, 30) == 1)
	 {
	 on_event(&event);
	 }
#endif

      update();

      if (redraw)
	 render();
      }

   cleanup();

   return 0;
   }

////////////////////////////////////////////////////////////////////////////////
// Sample callback from
// http://people.freedesktop.org/~idr/OpenGL_tutorials/01-SDL-intro.html

#ifdef USE_SDL_WAITEVENT
Uint32 App::timer_callback(Uint32 interval, void *param)
   {
   SDL_Event e;
 
   e.type = SDL_USEREVENT;
   e.user.code = 0;
   e.user.data1 = NULL;
   e.user.data2 = NULL;
   SDL_PushEvent(& e);
 
   return interval;
   }
#endif

////////////////////////////////////////////////////////////////////////////////
bool App::init()
   {
   if ((SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER)) == -1)
      {
      std::cerr << "SDL_Init() failed: " << SDL_GetError() << std::endl;
      return false;
      }

   if ((surf_display = SDL_SetVideoMode(Options::width, 
					Options::height, 
					Options::bpp, 
					SDL_HWSURFACE | SDL_DOUBLEBUF)) == NULL)
      {
      std::cerr << "SDL_SetVideoMode() failed: " 
		<< SDL_GetError() << std::endl;
      return false;
      }

   SDL_EnableKeyRepeat(500, 100);


   if (!create_new_surface(surf_selection)) { return false; }
   if (!create_new_surface(surf_fractal))   { return false; }

   fractal = new Mandelbrot();
   fractal->init(surf_fractal);
   can_set_julia_k = true;
   redraw = true;

   set_caption();

#ifdef USE_SDL_WAITEVENT
   SDL_TimerID timer_id = SDL_AddTimer(10, App::timer_callback, NULL);
#endif

   return true;
   }

////////////////////////////////////////////////////////////////////////////////

void App::update()
   {
   }

////////////////////////////////////////////////////////////////////////////////

void App::render()
   {
   if (redraw)
      {
      // Start with a clean slate.
      SDL_FillRect(surf_display, &surf_display->clip_rect, 
		   SDL_MapRGB(surf_display->format, 0, 0, 255));

      fractal->calc_set();
      SDL_BlitSurface(surf_fractal, NULL, surf_display, NULL);

      // draw a selection box
      if (setting_zoom)
	 {
	 SDL_BlitSurface(surf_selection, NULL, surf_display, NULL);
	 }

      // Show it to the user.
      SDL_Flip(surf_display);

      redraw = false;
      }
   }

////////////////////////////////////////////////////////////////////////////////

void App::cleanup()
   {

   if (surf_selection)
      {
      SDL_FreeSurface(surf_selection);
      surf_selection = NULL;
      }
   if (surf_display)
      {
      SDL_FreeSurface(surf_display);
      surf_display = NULL;
      }

   SDL_Quit();
   }

////////////////////////////////////////////////////////////////////////////////

void App::on_event(SDL_Event * event)
   {
   EventHandler::on_event(event);
   }

////////////////////////////////////////////////////////////////////////////////

void App::on_KeyDown(SDLKey sym, SDLMod mod, Uint16 unicode)
   {
   switch (sym)
      {
      case SDLK_ESCAPE:
      case SDLK_q:
	 SDL_Event event;
	 event.type = SDL_QUIT;
	 SDL_PushEvent(&event);
	 break;

      case SDLK_b:
      	 restore_view();
	 fractal->set_calc_needed();
	 redraw = true;
	 set_caption();
      	 break;

      case SDLK_s:
      	 save_image();
      	 break;

      case SDLK_UP:
	 fractal->inc_max_iter();
	 set_caption();
	 break;
      case SDLK_DOWN:
	 fractal->dec_max_iter();
	 set_caption();
	 break;


      case SDLK_m:
	 if (fractal)
	    {
	    delete fractal;
	    fractal = new Mandelbrot();
	    fractal->init(surf_fractal);
	    fractal->set_calc_needed();
	    redraw = true;
	    set_caption();
	    can_set_julia_k = true;
	    }
	 break;
      case SDLK_j:
	 if (fractal)
	    {
	    delete fractal;
	    fractal = new Julia();
	    fractal->init(surf_fractal);
	    fractal->set_calc_needed();
	    redraw = true;
	    set_caption();
	    can_set_julia_k = false;
	    }
	 break;

      case SDLK_k:
	 if (can_set_julia_k)
	    {
	    if (setting_julia_k) setting_julia_k = false;
	    else setting_julia_k = true;
	    }
	 break;

      default:
	 break;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_KeyUp(SDLKey sym, SDLMod mod, Uint16 unicode)
   {
   switch (sym)
      {
      case SDLK_UP:
	 fractal->set_calc_needed();
	 redraw = true;
	 break;
      case SDLK_DOWN:
	 fractal->set_calc_needed();
	 redraw = true;
	 break;
      }   
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_Exit()
   {
   running = false;
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_LButtonDown(int mx, int my)
   {
   if (setting_julia_k)
      {
      double K_Re = fractal->get_re_min() + mx * 
	    (fractal->get_size())/(Options::width-1);
      double K_Im = fractal->get_im_max() - my * 
	    (fractal->get_size())/(Options::height-1);

      delete fractal;
      fractal = new Julia();
      fractal->init(surf_fractal);
      ((Julia *)fractal)->set_k(K_Re, K_Im);
      fractal->set_calc_needed();
      redraw = true;

      set_caption();
      setting_julia_k = false;
      }
   else
      {
      selection_x = mx;
      selection_y = my;
      setting_zoom = true;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_LButtonUp(int mx, int my)
   {
   if (setting_zoom)
      {
      if ( (mx <= 0) || 
	   (mx >= Options::width - 1) || 
	   (my <= 0) || 
	   (my >= Options::height - 1))
	 {
	 // Selection cancelled
	 setting_zoom = false;
	 return;
	 }

      selection_width = mouse_move_width(mx, my);

      // calculate new min/max re/im
      double Re_scale = (fractal->get_size())/(Options::width-1);
      double Im_scale = (fractal->get_size())/(Options::height-1);

      double new_re_min = fractal->get_re_min() + (selection_x) * Re_scale;
      double new_im_max = fractal->get_im_max() - (selection_y) * Im_scale;
      double new_size = selection_width * Re_scale;

      
      view * v = new view;
      v->re_min = fractal->get_re_min();
      v->im_max = fractal->get_im_max();
      v->size = fractal->get_size();
      old_views.push_back(v);

      fractal->set_re_min(new_re_min);
      fractal->set_im_max(new_im_max);
      fractal->set_size(new_size);
      fractal->set_calc_needed();
      redraw = true;

      set_caption();

      setting_zoom = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_MouseMove(int mx, int my, 
		       int relx, int rely, 
		       bool left, bool right, bool middle)
   {
   if (setting_zoom)
      {
      Uint32 clear = SDL_MapRGBA(surf_selection->format, 
				0, 0, 0, SDL_ALPHA_TRANSPARENT);
      Uint32 white = SDL_MapRGBA(surf_selection->format, 
				 255, 255, 255, SDL_ALPHA_OPAQUE);

      SDL_FillRect(surf_selection, &surf_selection->clip_rect, clear);

      SDL_Rect rect;
      rect.x = selection_x;
      rect.y = selection_y;
      int w = mouse_move_width(mx, my);
      rect.w = rect.h = w;

      SDL_FillRect(surf_selection, &rect, white);

      rect.x = selection_x + 1;
      rect.y = selection_y + 1;
      rect.w = rect.h = w - 2;

      SDL_FillRect(surf_selection, &rect, clear);

      redraw = true;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_Resize(int w, int h)
   {
   redraw = true;
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_Expose()
   {
   redraw = true;
   }


////////////////////////////////////////////////////////////////////////////////
int App::mouse_move_width(int mx, int my)
   {
   int move_width = (mx - selection_x);
   if (mx > selection_x)
      {
      move_width *= -1;
      }
   
   int tmp_i = (my - selection_y);
   if (my < selection_y)
      {
      tmp_i *= -1;
      }
   
   if (tmp_i > move_width)
      move_width = tmp_i;

   return move_width;
   }


////////////////////////////////////////////////////////////////////////////////
bool App::create_new_surface(SDL_Surface * &surface)
   {
   if (surface)
      {
      SDL_FreeSurface(surface);
      surface = NULL;
      }

   SDL_Surface * tmp;
   if ((tmp = SDL_CreateRGBSurface(SDL_HWSURFACE|SDL_SRCALPHA, 
				   Options::width,
				   Options::height, Options::bpp, 0, 0, 0, 0))
       == NULL)
      {
      std::cerr << "SDL_CreateRGBSurface() failed in "
		<< "App::create_new_surface(): " 
		<< SDL_GetError() << std::endl;
      return false;
      }

   surface = SDL_DisplayFormatAlpha(tmp);
   SDL_FreeSurface(tmp);

   return true;
   }


////////////////////////////////////////////////////////////////////////////////

void App::restore_view()
   {
   if (old_views.size() > 0)
      {
      view * v = old_views.back();
      old_views.pop_back();
      fractal->set_re_min(v->re_min);
      fractal->set_im_max(v->im_max);
      fractal->set_size(v->size);
      delete v;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::set_caption()
   {
   std::stringstream ss;
   ss << fractal->get_name() << " / "
      << "Real: " << fractal->get_re_min()
       << " to "<< fractal->get_re_min() + fractal->get_size()
      << " / Imaginary: " << fractal->get_im_max() - fractal->get_size()
      << " to " << fractal->get_im_max()
      << " / Max iterations: " << fractal->get_max_iter();
   SDL_WM_SetCaption(ss.str().c_str(), NULL);
   }


////////////////////////////////////////////////////////////////////////////////
// This file name generation is kuldgy.  Need to learn how to do file i/o for 
// real!

int scandir_filter(const struct dirent * d)
   {
   if (memcmp(d->d_name, "mandelbrot-", 11) != 0) return 0;
   if (memcmp(d->d_name+14, ".bmp", 4) != 0) return 0;
   if (isdigit(d->d_name[11]) 
       && isdigit(d->d_name[12]) 
       && isdigit(d->d_name[13])) 
      return 1;
   return 0;
   }


bool App::save_image()
   {
   static int i = 0;
   char name[20];

   if (i == 0)
      {
      struct dirent ** file_list;
      int num_files = scandir(".", &file_list, scandir_filter, alphasort);
      if (num_files != 0)
	 sscanf(file_list[num_files-1]->d_name, "mandelbrot-%03d.bmp", &i);
      }
   i++;
   sprintf(name, "mandelbrot-%03d.bmp", i);

   FILE * file;
   if ((file = fopen(name, "wb")) == NULL)
      {
      std::cerr << "fopen() failed in save_image() on file " 
		<< name << " for saving." << std::endl;
      return false;
      }

   if (SDL_SaveBMP(surf_display, name) != 0)
      {
      std::cerr << "Unable to save image file " << name << "" << std::endl;
      return false;
      }

   return true;
   }


////////////////////////////////////////////////////////////////////////////////
#ifdef SAVE_AS_PNG
void png_user_error(png_structp png, png_const_charp s);
void png_user_warning(png_structp png, png_const_charp s);

void png_user_warning(png_structp png, png_const_charp s)
   {
   std::cerr << "libpng: warning " << (char *) s << std::endl;
   }


////////////////////////////////////////////////////////////////////////////////
void png_user_error(png_structp png, png_const_charp s)
   {
   std::cerr << "libpng: error " << (char *) s << std::endl;
   }


////////////////////////////////////////////////////////////////////////////////

bool App::save_image()
   {
   static int i = 0;
   char name[20];

   i++;
   sprintf(name, "mandelbrot-%03d.png", i);

   png_save_surface(name, surf_display);

   FILE * file;
   if ((file = fopen(name, "wb")) == NULL)
      {
      std::cerr << "fopen() failed in save_image() on file " 
		<< name << " for saving." << std::endl;
      return false;
      }


   
   png_structp png;
   png_infop png_info;

   if ((png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, 
				      png_user_error, png_user_warning)) == NULL)
      {
      std::cerr << "png_create_write_struct() failed in save_image()"
		<< std::endl;
      fclose(file);
      return false;
      }


   if ((png_info = png_create_info_struct(png)) == NULL)
      {
      std::cerr << "png_create_info_struct() failed in save_image()"
		<< std::endl;
      fclose(file);
      return false;
      }

   if (setjmp(png_jmpbuf(png)))
      {
      std::cerr << "png_jmpbuf() failed in save_image()" << std::endl;
      png_destroy_write_struct(&png, &png_info);
      fclose(file);
      return false;
      }

   png_init_io(png, file);
   int colortype = PNG_COLOR_MASK_COLOR;
   png_set_IHDR(png, png_info, 
		surf_display->w, surf_display->h,
		8, colortype, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);

   png_write_info(png, png_info);
   png_set_packing(png);

   png_bytep * rows = (png_bytep *) malloc(sizeof(png_bytep)*surf_display->h);
   for (int i = 0; i < surf_display->h; i++)
      {
      rows[i] = (png_bytep) surf_display->pixels +
	    i * surf_display->pitch;
      }

   png_write_image(png, rows);
   png_write_end(png, png_info);

   free(rows);
   png_destroy_write_struct(&png, &png_info);
   fclose(file);

   return true;
   }
#endif
